/* -*- Mode: C; eval : (c-set-style "bsd"); -*- */

#include <stdio.h>

typedef enum { GTABLE_DEFAULT, GTABLE_LEFT, GTABLE_CENTER, GTABLE_RIGHT } 
               Gtable_halign;
typedef enum { GTABLE_VDEFAULT, GTABLE_TOP, GTABLE_VCENTER, GTABLE_BOTTOM } 
               Gtable_valign;
/* GTABLE_APPEND tells the GTABLE routines to append to a table. 
   This value will always be -1; a name is provided to improve readability 
   of code */
#define GTABLE_APPEND -1

/* Each table entry has this form */
typedef struct _gtentry {
    struct _gtentry *next_row, *next_col;
    char            *contents;
    int             n_contents;
    int             n_row, n_col;  /* Allows multiple row/column entries */
    int             est_width;     /* Estimated width (used by some methods) */
    Gtable_halign   halign;
    Gtable_valign   valign;
} Gtable_entry;

/* A table is made up of entries */
typedef struct {
    int          cookie;      /* Used to validate the structure */
    int          rows, cols;
    int          cur_row;     /* Current row, used for appending to a table */
    int          cur_col;     /* Current col, ditto */
    Gtable_entry *first;      /* Upper left corner */
} Gtable;

/* Output is managed by a method that knows how to output a table */
typedef int (*Gtable_row_fcn)( Gtable *, Gtable_entry *, int, void * );
typedef int (*Gtable_col_fcn)( Gtable *, Gtable_entry *, int, int, void * );
typedef int (*Gtable_fcn)( Gtable *, void * );
typedef struct _gtable_output {
    int (*start_table)( Gtable *, struct _gtable_output *, void * );
    Gtable_fcn     end_table;
    Gtable_row_fcn start_row;
    Gtable_row_fcn end_row;
    Gtable_col_fcn output_col;
    Gtable_col_fcn output_multicol;
    void (*free_fcn)( struct _gtable_output * );
    void *extra_data;
} Gtable_output;

/* Function defintions */

/* &ctx, nrow, ncol (optional) */
Gtable *GtableCreate( int, int );
/* ctx, row, col, string */
int GtableAddEntry( Gtable *, int, int, const char * );
/* ctx, row, col, fmt, (sprintf-style-args) */
int GtableAddEntryFmt( Gtable *, int, int, const char *, ... );
/* ctx, row, col, nrow, ncol, string */
int GtableAddBlock( Gtable *, int, int, int, int, const char * );
/* ctx, row, sepchar, string - defines a whole row at once */
int GtableAddRow( Gtable *, int, char, const char * );
/* ctx */
int GtableNewRow( Gtable * );
/* ctx, method, extra_data */
int GtableOutput( Gtable *, Gtable_output *, void * );
/* &ctx */
int GtableFree( Gtable ** );

/* Method definitions */
Gtable_output *Gtable_CreateText( FILE *, const char * );
Gtable_output *Gtable_CreateHTML( FILE *, const char * );
Gtable_output *Gtable_CreateXML( FILE *, const char * );
int Gtable_XML_Set_col( Gtable_output *, int, const char * );
int Gtable_XML_Set_row( Gtable_output *, int, const char *, const char * );
Gtable_output *Gtable_CreateLaTeX( FILE *, const char * );
Gtable_output *Gtable_CreateTab( FILE *, const char * );
void GtableOutputFree( Gtable_output * );

/* Special routines for XML output */
int Gtable_XML_Names( Gtable_output *, const char *,
		      const char *, const char *, const char * );
int Gtable_XML_Column_names( Gtable *, Gtable_output *, int, const char * );

/*
 *  Not yet done: HDF
 */

/* Unresolved.  Where does alignment go?  In the creation routine (like LaTeX)?
   In the AddEntry routine (more like HTML, though could fake out LaTeX)?
  
   Is there a raw data representation?  E.g., bytes for double data, with
   char for a string rep (allows a Gtable_Binary output method).
 
   Controlling width (or height) of entries.  Is there something like latex 
   p{3in} ?
   Numbers are a special case; they are often better presented if the decimal
   points line up.  Is this a special number formatting case?  Should there 
   be a hook for a formatting function?

   Decoration like cell borders and LaTeX \hline and |?  Colors?
   (Things like colors should probably be a property of the output method)

   Headings for columns and rows?  

   Note that the current set of output methods are really just special cases
   of a single method that has strings for the start/end table, row, and 
   column formatting, along with a special preface that can compute things 
   like column widths (for text output)
 */
