#include <stdio.h>
#include "gtable.h"
#include "gtableimpl.h"
#include <stdlib.h>
#include <string.h>

/*
 * Simple LaTeX output.
 */

typedef struct {
    FILE *file;
    const char *newline;
} glatex_data_t;

static int gi_nop( Gtable *ctx )
{
    return 0;
}

/* Currently, we only support center formatting */
static int gi_start_table( Gtable *ctx, Gtable_output *output, 
			  void *extra_data )
{
    glatex_data_t *gt = (glatex_data_t *)(output->extra_data);
    int i;
    fputs( "\\begin{tabular}[", gt->file );
    for (i=0; i<ctx->cols; i++) {
	fputs( "c", gt->file );
    }
    fputs( "]", gt->file );
    fputs( gt->newline, gt->file );
    return 0;
}

static int gi_end_table( Gtable *ctx, void *extra_data )
{
    glatex_data_t *gt = (glatex_data_t *)extra_data;
    fputs( "\\end{tabular}", gt->file );
    fputs( gt->newline, gt->file );
    return 0;
}

static int gi_start_row( Gtable *ctx, Gtable_entry *row, int row_num, 
		      void *extra_data )
{
/*    glatex_data_t *gt = (glatex_data_t *)extra_data;
      fputs( "", gt->file ); */
    return 0;
}
static int gi_end_row( Gtable *ctx, Gtable_entry *row, int row_num, 
		      void *extra_data )
{
    glatex_data_t *gt = (glatex_data_t *)extra_data;
    fputs( "\\\\", gt->file );
    fputs( gt->newline, gt->file );
    return 0;
}

static int gi_output_col( Gtable *ctx, Gtable_entry *col, 
			 int row_num, int col_num, void *extra_data )
{
    glatex_data_t *gt = (glatex_data_t *)extra_data;
    const char *align_string, *valign_string;

    switch (col->halign) {
    case GTABLE_LEFT:
	align_string = "LEFT"; 	 break;
    case GTABLE_RIGHT:
	align_string = "RIGHT";	 break;
    case GTABLE_CENTER:
    case GTABLE_DEFAULT:
	align_string = "CENTER"; break;
    }

    switch (col->valign) {
    case GTABLE_TOP:
	valign_string = "TOP"; break;
    case GTABLE_BOTTOM:
	valign_string = "BOTTOM"; break;
    case GTABLE_VDEFAULT:
    case GTABLE_VCENTER:
	valign_string = "CENTER"; break;
    }

    if (col_num > 0) 
	fputs( "&", gt->file );
    fputs( col->contents ? col->contents : "", gt->file );
    return 0;
}

static void gi_free( Gtable_output *output )
{
    free( output->extra_data );
    free( output );
}

/* 
   Nulls for file and newline give the defaults (stdout and \n).
   setting newline to "\r\n" makes the output easier to read on Windows 
   systems.
*/
Gtable_output *Gtable_CreateLaTeX( FILE *file, const char *newline )
{
    Gtable_output *new;
    glatex_data_t  *gt;

    new = (Gtable_output *)malloc( sizeof(Gtable_output) );
    if (!new) return NULL;

    new->start_table     = gi_start_table;
    new->end_table       = gi_end_table;

    new->start_row       = gi_start_row;
    new->end_row         = gi_end_row;
    new->output_col      = gi_output_col;
    new->output_multicol = (Gtable_col_fcn)gi_nop;
    new->free_fcn        = gi_free;

    gt                   = (glatex_data_t *)malloc( sizeof(glatex_data_t) );
    if (!gt) return 0;
    gt->file             = file ? file : stdout;
    gt->newline          = newline ? strdup(newline) : strdup( "\n" );
    new->extra_data      = gt;

    return new;
}

