#include <stdio.h>
#include "gtable.h"
#include "gtableimpl.h"
#include <stdlib.h>
#include <string.h>

/*
 * Simple TAB-separated output, suitable for input into spreadsheets.
 */

typedef struct {
    FILE *file;
    const char *newline;
} gtab_data_t;

static int gi_nop( Gtable *ctx )
{
    return 0;
}

static int gi_start_table( Gtable *ctx, Gtable_output *output, 
			  void *extra_data )
{
    return 0;
}

static int gi_end_table( Gtable *ctx, void *extra_data )
{
    return 0;
}

static int gi_start_row( Gtable *ctx, Gtable_entry *row, int row_num, 
		      void *extra_data )
{
    return 0;
}
static int gi_end_row( Gtable *ctx, Gtable_entry *row, int row_num, 
		      void *extra_data )
{
    gtab_data_t *gt = (gtab_data_t *)extra_data;
    fputs( gt->newline, gt->file );
    return 0;
}

static int gi_output_col( Gtable *ctx, Gtable_entry *col, 
			 int row_num, int col_num, void *extra_data )
{
    gtab_data_t *gt = (gtab_data_t *)extra_data;

    if (col_num > 0) 
	fputs( "\t", gt->file );
    fputs( col->contents ? col->contents : "", gt->file );
    return 0;
}

static void gi_free( Gtable_output *output )
{
    free( output->extra_data );
    free( output );
}

/* 
   Nulls for file and newline give the defaults (stdout and \n).
   setting newline to "\r\n" makes the output easier to read on Windows 
   systems.
*/
Gtable_output *Gtable_CreateTab( FILE *file, const char *newline )
{
    Gtable_output *new;
    gtab_data_t  *gt;

    new = (Gtable_output *)malloc( sizeof(Gtable_output) );
    if (!new) return NULL;

    new->start_table     = gi_start_table;
    new->end_table       = gi_end_table;

    new->start_row       = gi_start_row;
    new->end_row         = gi_end_row;
    new->output_col      = gi_output_col;
    new->output_multicol = (Gtable_col_fcn)gi_nop;
    new->free_fcn        = gi_free;

    gt                   = (gtab_data_t *)malloc( sizeof(gtab_data_t) );
    if (!gt) return 0;
    gt->file             = file ? file : stdout;
    gt->newline          = newline ? strdup(newline) : strdup( "\n" );
    new->extra_data      = gt;

    return new;
}

