#include <stdio.h>
#include "gtable.h"
#include "gtableimpl.h"
#include <stdlib.h>
#include <string.h>

typedef struct {
    FILE *file;
    const char *newline;
    int   *col_widths;
} gtext_data_t;

static int gi_nop( Gtable *ctx )
{
    return 0;
}

static int giend_row( Gtable *ctx, Gtable_entry *row, int row_num, 
		      void *extra_data )
{
    gtext_data_t *gt = (gtext_data_t *)extra_data;
    fputs( gt->newline, gt->file );
    return 0;
}

static int gioutput_col( Gtable *ctx, Gtable_entry *col, 
			 int row_num, int col_num, void *extra_data )
{
    gtext_data_t *gt = (gtext_data_t *)extra_data;
    int *col_widths = gt->col_widths;
    int i;
    char *string;

/*    printf ("(%d,%d) width is %d\n", row_num, col_num, col_widths[col_num] ); */
    /* Sanity check */
    if (col->est_width > col_widths[col_num]) {
	fprintf( stderr, 
 "Internal error: Actual column width (%d) larger than expected (%d)\n",
		 col->n_contents, col_widths[col_num] );
    }
    /* find the corresponding row/column and compare col widths */

    string = col->contents;
    if (!string) string = "";
    switch (col->halign) {
    case GTABLE_LEFT:
	fputs( string, gt->file );
	for (i=col->est_width; i<col_widths[col_num]; i++) 
	    fputc( ' ', stdout );
	break;

    case GTABLE_RIGHT:
	for (i=col->est_width; i<col_widths[col_num]; i++) 
	    fputc( ' ', stdout );
	fputs( string, gt->file );
	break;

    case GTABLE_CENTER:
    case GTABLE_DEFAULT:
    {
	int n_blank = col_widths[col_num] - col->est_width;
	int nb2 = n_blank / 2;
	
	for (i=0; i<nb2; i++) 
	    fputc( ' ', gt->file );
	fputs( string, gt->file );
	nb2 = n_blank - nb2;
	for (i=0; i<nb2; i ++) 
	    fputc( ' ', gt->file );
    }
    break;
    }
    return 0;
}

static int gi_findwidths( Gtable *ctx, Gtable_output *method, 
			  void *extra_data )
{
    Gtable_entry *row, *col;
    int          col_num, i;
    int          *col_widths;

    /* Setup the col widths array */
    col_widths = (int *)malloc( ctx->cols * sizeof(int) );
    for (i=0; i<ctx->cols; i++) 
	col_widths[i] = 0;

    row = ctx->first;
    while (row) {
	col = row;
	col_num = 0;
	while (col) {
	    if (col->est_width <= 0 && col->contents) {
		col->est_width = strlen( col->contents );
	    }
	    if (col->est_width > col_widths[col_num]) 
		col_widths[col_num] = col->est_width;
	    col_num++;
	    col = col->next_col;
	}
	row = row->next_row;
    }
    ((gtext_data_t*)(method->extra_data))->col_widths = col_widths;
    return 0;
}

static void gi_free( Gtable_output *output )
{
    gtext_data_t *gt = (gtext_data_t *)output->extra_data;
    if (gt) {
	if (gt->col_widths) free( gt->col_widths );
	free( gt );
    }
    free( output );
}

Gtable_output *Gtable_CreateText( FILE *file, const char *newline )
{
    Gtable_output *new;
    gtext_data_t  *gt;

    new = (Gtable_output *)malloc( sizeof(Gtable_output) );
    if (!new) return NULL;

    new->start_table     = gi_findwidths;
    new->end_table       = (Gtable_fcn)gi_nop;

    new->start_row       = (Gtable_row_fcn)gi_nop;
    new->end_row         = giend_row;
    new->output_col      = gioutput_col;
    new->output_multicol = (Gtable_col_fcn)gi_nop;
    new->free_fcn        = gi_free;

    gt                   = (gtext_data_t *)malloc( sizeof(gtext_data_t) );
    if (!gt) return 0;
    gt->file             = file ? file : stdout;
    gt->newline          = newline ? strdup(newline) : strdup( "\n" );
    gt->col_widths       = 0;
    new->extra_data      = gt;

    return new;
}
