#include "gtable.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Simple program to read input and generate a table in various formats,
 * using the gtable code.  
 *
 * ttable [-format=name] [-sep=char] <input >output
 * where name is one of: html, text, xml, latex
 * and sep is a separator character
 *
 * The XML format is very simple
 */

#define MAX_BUF 512
void Abort( const char * );

int main( int argc, char *argv[] )
{
    char buf[MAX_BUF];
    int  row;
    Gtable *ctx;
    Gtable_output *table_method=0;
    char          sep_char = ':';

    /* Create a table */
    ctx = GtableCreate( -1, -1 );

    /* Determine the table format from the command line */
    while (--argc) {
	argv++;
	if (strncmp( *argv, "-format=", 8 ) == 0) {
	    char *p = (*argv) + 8;
	    if (strcmp( p, "text" ) == 0) {
		table_method = Gtable_CreateText( 0, 0 );
	    }
	    if (strcmp( p, "tab" ) == 0) {
		table_method = Gtable_CreateTab( 0, 0 );
	    }
	    else if (strcmp( p, "html" ) == 0) {
		table_method = Gtable_CreateHTML( 0, 0 );
	    }
	    else if (strcmp( p, "latex" ) == 0 ) {
		table_method = Gtable_CreateLaTeX( 0, 0 );
	    }
	    else if (strcmp( p, "xml" ) == 0) {
		int i;
		table_method = Gtable_CreateXML( 0, 0 );
		Gtable_XML_Names( table_method, "<XMLTABLE>", "</XMLTABLE>",
				  "\t<ROW>", "</ROW>" );
		for (i=0; i<10; i++) {
		    char col_fmt[64];
		    sprintf( col_fmt, "<COL%d>%%s</COL%d>", i, i );
		    Gtable_XML_Column_names( 0, table_method, i, col_fmt );
		}
	    }
	    else {
		fprintf( stderr, "Unrecognized format %s; text used\n", p );
	    }
	}
	else if (strncmp( *argv, "-sep=", 5 ) == 0) {
	    sep_char = (*argv)[5];
	}
	else {
	    /* unrecognized arg */
	    fprintf( stderr, "Unrecognized arg %s\n", *argv );
	    break;
	}
    }

    /* Read from stdin until eof */
    row = 0;
    while (fgets( buf, MAX_BUF, stdin )) {
	buf[strlen(buf)-1] = 0;
	if (GtableAddRow( ctx, row, sep_char, (const char *)buf )) {
	    Abort( "Error adding row" );
	}
	row++;
    }

    if (!table_method) table_method = Gtable_CreateText( 0, 0 );

    GtableOutput( ctx, table_method, 0 );
    GtableOutputFree( table_method );
    GtableFree( &ctx );
    
    return 0;
}

void Abort( const char *msg )
{
    fprintf( stderr, "Aborting: %s\n", msg );
    exit(1);
}
